﻿using System;
using System.IO;
using System.Collections.Generic;

using OpenTK;
using OpenTK.Graphics.OpenGL;
using OpenTK.Graphics;
using OpenTK.Input;

using SimpleScene; //SimpleScene by David W. Jeske. https://github.com/jeske/SimpleScene

using Valve.VR; //OpenVR by Valve. https://github.com/ValveSoftware/openvr

namespace openvrtest3
{

    class Game : GameWindow
    {
        const int WINDOWWIDTH = 800;
        const int WINDOWHEIGHT = 600;
        const float NEARPLANE = 0.01f;
        const float FARPLANE = 256.0f;

        uint[] framebuffer;
        uint[] colorRenderTarget;
        int framebufferWidth;
        int framebufferHeight;

        //VR.
        public CVRSystem hmd { get; private set; }
        public CVRCompositor compositor { get; private set; }
        private TrackedDevicePose_t[] devicePoses;
        Matrix4 trackedPose;
        Vector3 trackedPosition;        

        bool initialized;

        //Simplescene.
        SSScene scene;
        SSCamera[] camera;

        //Simplescene boilerplate.
        protected SSMainShaderProgram mainShader;
        protected SSPssmShaderProgram pssmShader;
        protected SSInstanceShaderProgram instancingShader;
        protected SSInstancePssmShaderProgram instancingPssmShader;
        

        public Game() : base(WINDOWWIDTH, WINDOWHEIGHT, GraphicsMode.Default, "OpenTK OpenVR test")
        {
            if (!OpenVR.IsHmdPresent()) { Console.WriteLine("OPENVR: HMD is not present. Exiting."); return; }

            //Initialize openvr.
            EVRInitError peError = EVRInitError.None;

            OpenVR.Init(ref peError, EVRApplicationType.VRApplication_Scene);
            if (peError != EVRInitError.None) { Console.WriteLine("OPENVR: initialization failed: {0}", peError); return; }

            OpenVR.GetGenericInterface(OpenVR.IVRCompositor_Version, ref peError);
            if (peError != EVRInitError.None) { Console.WriteLine("OPENVR: compositor initialization failed: {0}", peError); return; }

            Console.WriteLine("OPENVR: initialized.");

            hmd = OpenVR.System;
            compositor = OpenVR.Compositor;

            devicePoses = new TrackedDevicePose_t[OpenVR.k_unMaxTrackedDeviceCount];

            this.Closing += new EventHandler<System.ComponentModel.CancelEventArgs>(Program_Closing);

            GenerateFrameBuffer();

            //SimpleScene setup.
            SSAssetManager.AddAssetArchive(new SSAssetArchiveHandler_FileSystem("./content"));

            mainShader = new SSMainShaderProgram(); // before mscene
            if (!mainShader.IsValid) { throw new Exception("Failed to build the main shader"); }
            
            pssmShader = new SSPssmShaderProgram();
            if (!pssmShader.IsValid) { pssmShader = null; }

            instancingShader = new SSInstanceShaderProgram();
            if (!instancingShader.IsValid) { instancingShader = null; }
            else { instancingShader.debugLocations(); }

            instancingPssmShader = new SSInstancePssmShaderProgram();
            if (!instancingPssmShader.IsValid) { instancingPssmShader = null; }



            scene = new SSScene(mainShader, pssmShader, instancingShader, instancingPssmShader);
            


            //barrel model.
            var mesh = SSAssetManager.GetInstance<SSMesh_wfOBJ>("./barrel/barrel.obj");
            SSObject droneObj = new SSObjectMesh(mesh);
            scene.AddObject(droneObj);
            droneObj.renderState.lighted = false;
            droneObj.Pos = new Vector3(0.1f, 1.2f, -0.1f);


            camera = new SSCamera[]
            {
                new SSCamera(),
                new SSCamera()
            };

            
            initialized = true;
        }


        #region FRAME BUFFER
        private void GenerateFrameBuffer()
        {
            uint x = 0;
            uint y = 0;
            hmd.GetRecommendedRenderTargetSize(ref x, ref y);
            framebufferWidth = (int)x;
            framebufferHeight = (int)y;

            framebuffer = new uint[2];
            GL.GenFramebuffers(2, framebuffer);

            colorRenderTarget = new uint[2];
            uint[] depthRenderTarget = new uint[2];

            GL.GenTextures(2, colorRenderTarget);
            GL.GenTextures(2, depthRenderTarget);

            int clamp = (int)OpenTK.Graphics.OpenGL.All.ClampToEdge;
            int nearest = (int)OpenTK.Graphics.OpenGL.All.Nearest;
            int linear = (int)OpenTK.Graphics.OpenGL.All.Linear;

            for (int eye = 0; eye < 2; ++eye)
            {
                GL.BindTexture(TextureTarget.Texture2D, colorRenderTarget[eye]);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, ref clamp);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, ref clamp);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, ref nearest);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, ref nearest);
                GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba8, framebufferWidth, framebufferHeight, 0, PixelFormat.Rgba, PixelType.UnsignedByte, IntPtr.Zero);

                GL.BindTexture(TextureTarget.Texture2D, depthRenderTarget[eye]);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, ref linear);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, ref linear);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, ref clamp);
                GL.TexParameterI(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, ref clamp);
                GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.DepthComponent24, framebufferWidth, framebufferHeight, 0, PixelFormat.DepthComponent, PixelType.UnsignedInt, IntPtr.Zero);

                GL.BindFramebuffer(FramebufferTarget.Framebuffer, framebuffer[eye]);
                GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, colorRenderTarget[eye], 0);
                GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.DepthAttachment, TextureTarget.Texture2D, depthRenderTarget[eye], 0);
            }


            GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
        }
        #endregion


        #region BOILERPLATE

        void Program_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            OpenVR.Shutdown();
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);
            GL.Enable(EnableCap.DepthTest);
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);

            GL.Viewport(this.X, this.Y, framebufferWidth, framebufferHeight);

            float ratio = framebufferWidth / (float)framebufferHeight;

            Matrix4 projection = Matrix4.CreatePerspectiveFieldOfView((float)Math.PI / 4, ratio, NEARPLANE, FARPLANE);
            GL.MatrixMode(MatrixMode.Projection);
            GL.LoadMatrix(ref projection);
        }

        #endregion


        protected override void OnUpdateFrame(FrameEventArgs e)
        {
            if (!initialized)
                return;

            base.OnUpdateFrame(e);            
            scene.Update((float)e.Time);
        }

        private void UpdateTracking()
        {
            var err = OpenVR.Compositor.WaitGetPoses(devicePoses, new TrackedDevicePose_t[0]);
            if (err != EVRCompositorError.None) { Console.WriteLine("OPENVR: Failed to get poses: {0}", err); }

            TrackedDevicePose_t hmdPose = devicePoses[OpenVR.k_unTrackedDeviceIndex_Hmd];
            if (!hmdPose.bPoseIsValid) { Console.WriteLine("OPENVR: HMD pose is not valid."); }


            //update camera.
            trackedPose = devicePoses[OpenVR.k_unTrackedDeviceIndex_Hmd].mDeviceToAbsoluteTracking.ToGLMatrix4();
            trackedPosition = trackedPose.ExtractTranslation();


            for (int eye = 0; eye < 2; eye++)
            {
                camera[eye].updateMat(ref trackedPose);
                camera[eye].localMat = trackedPose;
                camera[eye].worldMat = trackedPose;

                Matrix4 eyePos = hmd.GetEyeToHeadTransform((EVREye)eye).ToGLMatrix4();
                Vector3 eyeOffset = eyePos.ExtractTranslation();

                camera[eye].Pos = trackedPosition + (camera[eye].Right * eyeOffset.X) +  (camera[eye].Up * eyeOffset.Y)  +  (camera[eye].Dir * eyeOffset.Z);
            }
        }


        private void DrawMesh(EVREye eye)
        {
            GL.Enable(EnableCap.CullFace);
            GL.CullFace(CullFaceMode.Back);
            GL.Enable(EnableCap.DepthTest);
            GL.Enable(EnableCap.DepthClamp);
            GL.DepthMask(true);

            //Matrix4 eyePos = hmd.GetEyeToHeadTransform((EVREye)eye).ToGLMatrix4();
            scene.renderConfig.invCameraViewMatrix = camera[(int)eye].localMat.Inverted();

            Matrix4 eyeProj = hmd.GetProjectionMatrix(eye, NEARPLANE, FARPLANE, EGraphicsAPIConvention.API_OpenGL).ToGLMatrix4();
            scene.renderConfig.projectionMatrix = eyeProj;

            scene.Render();
        }

        protected override void OnRenderFrame(FrameEventArgs e)
        {
            if (!initialized)
                return;

            UpdateTracking();

            base.OnRenderFrame(e);            

            for (int eye = 0; eye < 2; ++eye)
            {
                camera[eye].preRenderUpdate((float)e.Time);

                GL.BindFramebuffer(FramebufferTarget.Framebuffer, framebuffer[eye]);
                GL.Viewport(0, 0, framebufferWidth, framebufferHeight);

                GL.ClearColor(.5f, .2f, .0f, .0f);
                GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

                //Draw mesh.
                DrawMesh((EVREye)eye);
                
                //Submit to HMD.
                Texture_t tex = new Texture_t();
                tex.eColorSpace = EColorSpace.Gamma;
                tex.eType = EGraphicsAPIConvention.API_OpenGL;
                tex.handle = (IntPtr)colorRenderTarget[eye];
                VRTextureBounds_t bounds = new VRTextureBounds_t() { uMin = 0, vMin = 0, uMax = 1f, vMax = 1f };
                compositor.Submit((EVREye)eye, ref tex, ref bounds, EVRSubmitFlags.Submit_Default);
            }

            //Render to monitor.
            /*
            GL.BindFramebuffer(FramebufferTarget.Framebuffer, 0);
            GL.Viewport(0, 0, WINDOWWIDTH, WINDOWHEIGHT);
            GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
            DrawMesh(EVREye.Eye_Right);
            GL.BlitFramebuffer(0, 0, framebufferWidth, framebufferHeight, 0, 0, WINDOWWIDTH, WINDOWHEIGHT, ClearBufferMask.ColorBufferBit, BlitFramebufferFilter.Linear);
            GL.BindFramebuffer(FramebufferTarget.ReadFramebuffer, 0);
            SwapBuffers();
             */
        }

        static void Main(string[] args)
        {
            using (Game game = new Game())
            {
                game.Run(30, 90);
            }
        }        
    }

    struct FramebufferDesc
	{
		public uint m_nDepthBufferId;
        public uint m_nRenderTextureId;
        public uint m_nRenderFramebufferId;
        public uint m_nResolveTextureId;
        public uint m_nResolveFramebufferId;
	};
}
